@@ -1,8 +1,8 @@
 /*
 	tofrodos.c	Converts text files between DOS and Unix formats.
-	Copyright (c) 1996 by Christopher S L Heng. All rights reserved.
+	Copyright (c) 1996-1999 by Christopher S L Heng. All rights reserved.
 
-	$Id: tofrodos.c 1.2 1996/06/11 21:53:07 chris Exp $
+	$Id: tofrodos.c 1.7 1999/03/15 14:16:04 chris Exp $
 */
 
 /* this should always be first */
@@ -15,22 +15,35 @@
 #include <string.h>	/* strrchr(), strlen(), strcpy(), strcat() */
 #include <sys/stat.h>	/* stat() */
 
+#if defined(_MSC_VER) || defined(__WATCOMC__)
+#include <sys/utime.h>
+#else	/* everybody else keeps this in the include directory */
+#if defined(UNIX)
+#include <sys/types.h>
+#endif
+#include <utime.h>
+#endif
+
 /* conditionally included headers */
-#if defined(MSDOS)
+#if defined(MSDOS) || defined(WIN32)
 #include <fcntl.h>	/* O_BINARY */
-#include <io.h>		/* chmod(), setmode(), isatty() */
+#include <io.h>		/* chmod(), setmode(), isatty(), mktemp() (VC++) */
 #endif
 
 #if defined(HAVE_GETOPT_H)
 #include <getopt.h>	/* optind */
 #endif
 
-#if defined(HAVE_MKTEMP_H)
+#if defined(HAVE_MKTEMP_H) && defined(MKTEMP_HEADER)
 #include MKTEMP_HEADER
 #endif
 
 #if defined(HAVE_UNISTD_H)
-#include <unistd.h>	/* chmod(), mktemp(), isatty() */
+#include <unistd.h>	/* chmod(), mktemp(), isatty(), chown() */
+#endif
+
+#if defined(__DJGPP__) || defined(__GO32__)
+#include <dir.h>
 #endif
 
 /* our headers */
@@ -42,11 +55,12 @@
 /* macros */
 #define	BAKEXT		".bak"	/* backup file extension */
 #define	DIRSLASH	'/'	/* works for both MSDOS and Unix */
+#define	DIRSLASHSTR	"/"
 #define	MKTEMP_TEMPL	"XXXXXX"
 #define	NEWBUFSIZ	16384	/* buffer size for the files */
 
 /* conditional macros */
-#if defined(MSDOS)
+#if defined(MSDOS) || defined(WIN32)
 #if !defined(_MAX_DIR) || (_MAX_DIR < 260)	/* MAXDIRSIZE */
 #define MAXDIRSIZE	260
 #else
@@ -62,9 +76,12 @@
 #else
 #define	MAXPATHSIZE	_MAX_PATH
 #endif
+#if !defined(_MAX_DRIVE)
+#define	_MAX_DRIVE	3		/* for the benefit of djgpp */
+#endif
 #endif
 
-#if defined(MSDOS)
+#if defined(MSDOS) || defined(WIN32)
 #define	INFILEMODE	"rb"
 #define	OUTFILEMODE	"wb"
 #else
@@ -76,6 +93,7 @@
 #define	CURRENTDIR	"./"
 #endif
 
+
 /* global variables */
 int abortonerr ; /* 1 if should abort when there is error in any file */
 		/* in a list of files, 0 carry on (default) */
@@ -84,18 +102,24 @@ int alwaysconvert ; /* convert all \r\n to \r\r\n when direction */
 		/* DOSTOUNIX */
 int direction = DEFDIRECTION ; /* UNIXTODOS or DOSTOUNIX */
 int forcewrite ; /* convert even if file is not writeable */
+int preserve ;	/* 1 if we are to preserve owner (Unix) and date (all) */
 char * progname = VERSN_PROGNAME ;/* name of binary (ie, argv[0]) */
 int overwrite = 1 ;	/* 1 = overwrite original file, 0 = make backup */
 int verbose ;
 
 /* local variables */
 static char * infilename = "stdin" ;
-static unsigned short origfilemode ;	/* file mode of original file */
 static FILE * tempfp ;
 static char * tempfilename ;
 
 /* local functions */
-static int checkmode ( char * filename );
+#if !defined(UNIX)
+static int checkmode ( char * filename, unsigned short * origfilemodep,
+	struct utimbuf * filetimebufp );
+#else
+static int checkmode ( char * filename, unsigned short * origfilemodep,
+	struct utimbuf * filetimebufp, uid_t * ownerp, gid_t * groupp );
+#endif
 static int convert ( FILE * infp, FILE * outfp );
 static int openandconvert ( char * filename );
 
@@ -126,9 +150,9 @@ int main ( int argc, char ** argv )
 		return EXIT_ERROR ;
 	    }
 	    /* otherwise stdin has been redirected */
-#if defined(MSDOS)
+#if defined(MSDOS) || defined(WIN32)
 	    /* need to make sure the input and output files are binary */
-	    /* on MSDOS */
+	    /* on MSDOS and WIN32 */
 	    setmode( fileno( stdin ), O_BINARY );
 	    setmode( fileno( stdout ), O_BINARY );
 #endif
@@ -194,7 +218,7 @@ void sighandler ( int sig )
 
 	At the same time we also save the current mode of the file
 	so that we can set the converted file to the same mode. The
-	value is saved in origfilemode.
+	value is saved in the variable pointed to by origfilemodep.
 
 	Returns: 0 on success, -1 on error.
 
@@ -205,7 +229,13 @@ void sighandler ( int sig )
 	3) the file is not writeable and forcewrite is zero.
 	An error message is displayed on error.
 */
-static int checkmode ( char * filename )
+#if !defined(UNIX)
+static int checkmode ( char * filename, unsigned short * origfilemodep,
+	struct utimbuf * filetimebufp )
+#else
+static int checkmode ( char * filename, unsigned short * origfilemodep,
+	struct utimbuf * filetimebufp, uid_t * ownerp, gid_t * groupp )
+#endif
 {
 	struct stat statbuf ;
 
@@ -216,15 +246,23 @@ static int checkmode ( char * filename )
 		return -1 ;
 	}
 	/* save the mode */
-	origfilemode = statbuf.st_mode ;
+	*origfilemodep = statbuf.st_mode ;
+	/* save the file times for restore later */
+	filetimebufp->actime = statbuf.st_atime ;
+	filetimebufp->modtime = statbuf.st_mtime ;
+#if defined(UNIX)
+	/* save the owner and group id */
+	*ownerp = statbuf.st_uid ;
+	*groupp = statbuf.st_gid ;
+#endif
 	/* check if file can be read - this is actually redundant for */
 	/* DOS systems. */
-	if (!(origfilemode & S_IRUSR)) { /* not readable */
+	if (!(statbuf.st_mode & S_IRUSR)) { /* not readable */
 		fprintf( stdout, EMSG_NOTREADABLE, progname, filename );
 		return -1 ;
 	}
 	/* check if file can be written to, if forcewrite is 0 */
-	if (!forcewrite && !(origfilemode & S_IWUSR)) { /* not writeable */
+	if (!forcewrite && !(statbuf.st_mode & S_IWUSR)) { /* not writeable */
 		fprintf( stdout, EMSG_NOTWRITEABLE, progname, filename );
 		return -1 ;
 	}
@@ -252,7 +290,7 @@ static int convert ( FILE * infp, FILE * outfp )
 	int prevch ;
 	int c ;
 
-	/* actually it is very simple to do the conversion in DOS */
+	/* actually it is very simple to do the conversion in DOS/WIN32 */
 	/* because the stdio library does this work automatically for */
 	/* us. But because we want this program to work on Linux as */
 	/* well, a little bit of work stands before us (but only a little). */
@@ -266,36 +304,48 @@ static int convert ( FILE * infp, FILE * outfp )
 		/* text file with lines that have a '\r' just */
 		/* before a '\n'. These lines will */
 		/* not be converted to "\r\r\n" since the function */
-                /* below assumes the line has already been converted. */
-                /* To force the conversion of all \n to \r\n regardless */
+		/* below assumes the line has already been converted. */
+		/* To force the conversion of all \n to \r\n regardless */
 		/* of preceding characters, set alwaysconvert to 1. */
 		while ( (c = getc( infp )) != EOF ) {
-			if (c == '\n' && (!alwaysconvert && prevch != '\r')) {
+			if (c == '\n' && (alwaysconvert || prevch != '\r')) {
 				if (putc( '\r', outfp ) == EOF)
 					break ;
 			}
-                        /* always emit the current character */
+#if defined(__BORLANDC__)
+#pragma warn -sig
+#endif
+			/* always emit the current character */
 			if (putc( c, outfp ) == EOF)
-                        	break ;
+				break ;
+#if defined(__BORLANDC__)
+#pragma warn .sig
+#endif
 			prevch = c ;                        	 
-                }
+		}
 	}
 	else if (direction == DOSTOUNIX) {
 	    if (!alwaysconvert) {
 		/* basically we withhold emitting any '\r' until we */
 		/* are sure that the next character is not a '\n'. */
 		/* If it is not, we emit the '\r', if it is, we */
-                /* only emit the '\n'. */
+		/* only emit the '\n'. */
 		while ( (c = getc( infp )) != EOF ) {
 			if (prevch == '\r') {
 				/* '\r' is a special case because we don't */
 				/* emit a '\r' until the next character */
 				/* has been read */
 				if (c == '\n') { /* a "\r\n" pair */
+#if defined(__BORLANDC__)
+#pragma warn -sig
+#endif
 					/* discard previous '\r' and */
 					/* just put the '\n' */
 					if (putc( c, outfp ) == EOF)
 						break ;
+#if defined(__BORLANDC__)
+#pragma warn .sig
+#endif
 				}
 				else {	/* prevch was a standalone '\r' */
                                 	/* emit the standalone '\r' */
@@ -303,18 +353,30 @@ static int convert ( FILE * infp, FILE * outfp )
 						break ;
 					/* emit the current character if */
                                         /* it is not a '\r' */
+#if defined(__BORLANDC__)
+#pragma warn -sig
+#endif
 					if (c != '\r') {
 						if (putc( c, outfp ) == EOF)
 							break ;
-                                        }
+					}
+#if defined(__BORLANDC__)
+#pragma warn .sig
+#endif
 				}
 			}
 			else { /* prevch was not '\r' */
 				/* emit current character if it is not */
 				/* a '\r' */
 				if (c != '\r') {
+#if defined(__BORLANDC__)
+#pragma warn -sig
+#endif
 					if (putc( c, outfp ) == EOF)
 						break ;
+#if defined(__BORLANDC__)
+#pragma warn .sig
+#endif
 				}
 			}
 			prevch = c ;
@@ -323,8 +385,14 @@ static int convert ( FILE * infp, FILE * outfp )
 	    else { /* eliminate all '\r' */
 		while ((c = getc( infp )) != EOF) {
 			if (c != '\r') {
+#if defined(__BORLANDC__)
+#pragma warn -sig
+#endif
 				if (putc( c, outfp ) == EOF)
 					break ;
+#if defined(__BORLANDC__)
+#pragma warn .sig
+#endif
 			}
 			/* else skip all carriage returns */
 		}
@@ -356,24 +424,32 @@ static int convert ( FILE * infp, FILE * outfp )
 */
 static int openandconvert ( char * filename )
 {
-	FILE * infp ;
-        FILE * outfp ;
-	int err ;
-	char * bakfilename ;
-#if defined(MSDOS)
-	char drv[_MAX_DRIVE];
-	char dir[MAXDIRSIZE];
-	char fname[MAXFILESIZE];
-	char tempname[MAXPATHSIZE];
+	FILE *			infp ;
+	FILE *			outfp ;
+	int				err ;
+	char *			bakfilename ;
+	unsigned short	origfilemode ;	/* file mode of original file */
+	struct utimbuf	filetimebuf ;
+#if defined(MSDOS) || defined(WIN32)
+	char 			drv[_MAX_DRIVE];
+	char 			dir[MAXDIRSIZE];
+	char 			fname[MAXFILESIZE];
+	char 			tempname[MAXPATHSIZE];
 #else
-	char * s ;
-	char * t ;
-	size_t len ;
-	int replacech ;
+	char * 			s ;
+	char * 			t ;
+	char * 			p ;
+	size_t 			len ;
+	int				replacech ;
+	char			c ;
+#endif
+#if defined(UNIX)
+	uid_t			ownerid ;
+	gid_t			groupid ;
 #endif
 #if NEWBUFSIZ > BUFSIZ
-	char * inbufptr ;
-	char * outbufptr ;
+	char *			inbufptr ;
+	char *			outbufptr ;
 #endif
 
 	/* make sure we initialise */
@@ -381,71 +457,92 @@ static int openandconvert ( char * filename )
 
 	if (filename != NULL) { /* stdin is not redirected */
 
-	    /* check for appropriate permissions on the file */
-            /* also saves the mode in origfilemode */
-	    if (checkmode( filename ))
-            	return -1 ;
+		/* check for appropriate permissions on the file */
+		/* also saves the mode in origfilemode */
+#if !defined(UNIX) 
+		if (checkmode( filename, &origfilemode, &filetimebuf ))
+			return -1 ;
+#else
+		if (checkmode( filename, &origfilemode, &filetimebuf,
+			&ownerid, &groupid ))
+			return -1 ;
+#endif
 
 	    /* we need to create a temporary and backup filename (if */
 	    /* applicable) in the same directory */
 	    /* as our file. This is easy to do for DOS since we have the */
 	    /* _splitpath(), _makepath() functions. */
-#if defined(MSDOS)
+#if defined(MSDOS) || defined(WIN32)
 	    _splitpath( filename, drv, dir, fname, NULL );
 	    _makepath( tempname, drv, dir, MKTEMP_TEMPL, NULL );
 	    tempfilename = xstrdup( tempname );
 	    if (!overwrite) {
-		_makepath( tempname, drv, dir, fname, BAKEXT );
-		if (!strcmp( tempname, filename )) {
-			fprintf( stderr, EMSG_BAKFILENAME, filename );
-			goto err_freetempfn ;
-		}
-		bakfilename = xstrdup( tempname );
+			_makepath( tempname, drv, dir, fname, BAKEXT );
+			if (!strcmp( tempname, filename )) {
+				fprintf( stderr, EMSG_BAKFILENAME, filename );
+				goto err_freetempfn ;
+			}
+			bakfilename = xstrdup( tempname );
 	    }
-#else	/* not MSDOS */
+#else	/* neither MSDOS nor WIN32 */
 	    /* check if there is a path prefix */
 	    if ((s = strrchr( filename, DIRSLASH )) != NULL) {
-		*s = '\0';
-		replacech = 1 ;
-		len = strlen( filename ) ;
-		t = filename ;
+			c = *++s ;	/* save the character after the slash */
+			*s = '\0';
+			replacech = 1 ;
+			len = strlen( filename ) ;
+			t = filename ;
 	    }
 	    else {
-	    	replacech = 0 ;
+	    	replacech = c = 0 ;	/* I init c to suppress the warning */
+					/* issued by gcc -Wall */
 	    	len = sizeof(CURRENTDIR) - 1 ;
 	    	t = CURRENTDIR ;
 	    }
+	    /* got to add the extension manually */
 	    tempfilename = xmalloc( len + sizeof(MKTEMP_TEMPL) );
-	    strcpy( tempfilename, t );
-	    strcat( tempfilename, MKTEMP_TEMPL );
-	    if (!overwrite) {
-		/* for the backup file in Unix systems, just append a */
-		/* .bak to the existing filename */
-		bakfilename = xmalloc( len + sizeof(BAKEXT) );
-		strcpy( bakfilename, filename );
-		strcat( bakfilename, BAKEXT );
-	    }
+	    strcpy( tempfilename, t ); /* path leading to filename */
+	    strcat( tempfilename, MKTEMP_TEMPL ); /* filename */
 	    if (replacech)
-		*s = DIRSLASH ;
+		*s = c ;
+	    if (!overwrite) {
+			bakfilename	= xmalloc( strlen( filename ) +
+							sizeof( BAKEXT ) );
+			/* copy the filename */
+			strcpy( bakfilename, filename );
+			/* strip any trailing extension */
+			p = bakfilename + strlen( bakfilename ) - 1 ;
+			while (p != bakfilename) {
+				if (*p == '.') {
+					*p = '\0' ; /* strip current extension */
+					break ;
+				}
+				if (*p == '/') /* no extension */
+					break ;
+				p-- ;
+			}
+			/* add the new extension */
+			strcat( bakfilename, BAKEXT );
+	    }
 #endif
 	    /* make the temp filename */
 	    if (mktemp( tempfilename ) == NULL) {
-		fprintf( stderr, EMSG_NOTEMPNAME, progname );
+			fprintf( stderr, EMSG_NOTEMPNAME, progname );
 err_freebakfn:
-		if (!overwrite && bakfilename != NULL)
-			free( bakfilename );
-#if defined(MSDOS)
+			if (!overwrite && bakfilename != NULL)
+				free( bakfilename );
+#if defined(MSDOS) || defined(WIN32)
 err_freetempfn:
 #endif
-		free( tempfilename );
-		tempfilename = NULL ;
-		return -1 ;
+			free( tempfilename );
+			tempfilename = NULL ;
+			return -1 ;
 	    }
 
 	    /* open the filename as the input file */
 	    if ((infp = fopen( filename, INFILEMODE )) == NULL) {
-		fprintf( stderr, EMSG_OPENFILE, progname, filename );
-		goto err_freebakfn ;
+			fprintf( stderr, EMSG_OPENFILE, progname, filename );
+			goto err_freebakfn ;
 	    }
 	    /* associate the infilename with the filename for error */
 	    /* messages */
@@ -453,9 +550,9 @@ err_freetempfn:
 
 	    /* open the temp file as the output file */
 	    if ((tempfp = fopen( tempfilename, OUTFILEMODE )) == NULL) {
-		fprintf( stderr, EMSG_OPENFILE, progname, tempfilename );
-		fclose( infp );
-		goto err_freebakfn ;
+			fprintf( stderr, EMSG_OPENFILE, progname, tempfilename );
+			fclose( infp );
+			goto err_freebakfn ;
 	    }
 	    outfp = tempfp ;
 
@@ -463,11 +560,19 @@ err_freetempfn:
 	else { /* filename == NULL, ie stdin is redirected */
 		infp = stdin ;
 		outfp = stdout ;
+
+		/* not needed, but we do this for the record, and for */
+		/* fussy compilers */
+		origfilemode	= 0 ;
+		memset( &filetimebuf, 0, sizeof( struct utimbuf ) );
+
+#if defined(UNIX)
+		ownerid	= groupid	= 0 ;
+#endif
+
 	}
 
 #if NEWBUFSIZ > BUFSIZ
-	/* make sure that we have a big input and output buffer */
-	inbufptr = outbufptr = NULL ;
 	/* (don't use xmalloc() because it we can't get what we want, */
 	/* we just don't bother, and go ahead with the minimum) */
 	if ((inbufptr = malloc( NEWBUFSIZ )) != NULL)
@@ -499,29 +604,29 @@ err_freetempfn:
 	if (filename != NULL) { /* stdin was not redirected */
 
 	    if (err) { /* there was an error */
-            	/* delete the temp file since we've already created it */
-            	remove ( tempfilename );
-		goto err_freebakfn ;
+        	/* delete the temp file since we've already created it */
+        	remove ( tempfilename );
+			goto err_freebakfn ;
 	    }
 
 	    /* if we need to back up, delete any backup files of the */
 	    /* same name first (in case we are running on DOS where */
 	    /* a rename does not delete the file automatically) */
 	    if (!overwrite) {
-		/* remove existing backup files of same name */
-		remove( bakfilename );
-		/* rename the original file to the back up name */
-		rename( filename, bakfilename );
+			/* remove existing backup files of same name */
+			remove( bakfilename );
+			/* rename the original file to the back up name */
+			rename( filename, bakfilename );
 	    }
 	    else { /* if we do not need to back up delete the original file */
-#if defined(MSDOS)
-		/* for MSDOS, we need to make sure the file is writeable */
-		/* otherwise we cannot delete! Under Unix, this is not */
-		/* necessary, since a file can be deleted if we */
-		/* have write permissions for the directory */
-                chmod( filename, S_IRUSR|S_IWUSR );
+#if defined(MSDOS) || defined(WIN32)
+			/* for MSDOS, we need to make sure the file is writeable */
+			/* otherwise we cannot delete! Under Unix, this is not */
+			/* necessary, since a file can be deleted if we */
+			/* have write permissions for the directory */
+            chmod( filename, S_IRUSR|S_IWUSR );
 #endif
-		remove( filename );
+			remove( filename );
 	    }
 
 	    /* rename the temp file to the original file name */
@@ -533,10 +638,24 @@ err_freetempfn:
 
 	    /* free memory we allocated */
 	    if (!overwrite && bakfilename != NULL)
-		free( bakfilename );
+			free( bakfilename );
+
+		if (preserve) {
+			/* change to the original file time */
+			utime( filename, &filetimebuf );
+#if defined(UNIX)
+			/* Change the owner to the owner of the original file. */
+			/* We ignore errors since the user might simply want */
+			/* to use -p to set the file time, and not being root, */
+			/* chown() will fail on Linux. However, we issue an error */
+			/* message if the user wants verbosity. */
+			if (chown( filename, ownerid, groupid ) && verbose)
+				fprintf( stdout, EMSG_CHOWN, progname, filename );
+#endif
+		}
 
 	    /* change the file mode to reflect the original file mode */
-            chmod( filename, origfilemode );
+        chmod( filename, origfilemode );
 
 	}	/* stdin was not redirected */
 
